#include "pi.h"
#include "fft.h"
#include "cache.h"

#include <math.h>

static FFT_DATA_TYPE *SineTable=NULL;

#define TRIG_VARS                   \
 FFT_DATA_TYPE Nth_r,Nth_i,Pow_r,Pow_i;

#define INIT_TRIG(LENGTH)           \
 {size_t x=Log2(LENGTH)+1;          \
  F_Set(&Pow_r,1,0);                \
  F_Set(&Pow_i,0,0);                \
  Nth_i=SineTable[x];               \
  Nth_r=SineTable[x+1];             \
  F_Mul(&Nth_r,&Nth_r,&Nth_r);      \
  F_Add(&Nth_r,&Nth_r,&Nth_r);      \
  F_SetSign(&Nth_r,-1);             \
 }

#define NEXT_TRIG_POW               \
 {FFT_DATA_TYPE temp,T2;            \
   temp=Pow_r;                      \
   F_Mul(&Pow_r,&Pow_r,&Nth_r);     \
   F_Mul(&T2,&Pow_i,&Nth_i);        \
   F_Add(&Pow_r,&Pow_r,&temp);      \
   F_Sub(&Pow_r,&Pow_r,&T2);        \
   F_Mul(&T2,&Pow_i,&Nth_r);        \
   F_Add(&Pow_i,&Pow_i,&T2);        \
   F_Mul(&T2,&temp,&Nth_i);         \
   F_Add(&Pow_i,&Pow_i,&T2);        \
 }

// C=1, S=0 N1==N2
#define FHT_T1Butterfly(N1,N2,C,S)                \
 {size_t i1=N1,i2=N2;                             \
  FFT_DATA_TYPE cas1,temp;                        \
  cas1=Right[i1];                                 \
  temp=Left[i1];                                  \
  F_Add(&Left[i1],&temp,&cas1);                   \
  F_Sub(&Right[i2],&temp,&cas1);                  \
 }

#define FHT_T2Butterfly(N1,N2,C,S) \
 {FFT_DATA_TYPE Rx,Ri;             \
  size_t i1=N1,i2=N2;              \
  Rx=Right[i1];Ri=Right[i2];       \
  {FFT_DATA_TYPE cas1,Lx;          \
   F_Mul(&cas1,&Rx,&C);            \
   F_Mul(&Lx,&Ri,&S);              \
   F_Add(&cas1,&cas1,&Lx);         \
   Lx=Left[i1];                    \
   F_Add(&Left[i1],&Lx,&cas1);     \
   F_Sub(&Right[i1],&Lx,&cas1);    \
  }                                \
  {FFT_DATA_TYPE cas2,Li;          \
   F_Mul(&cas2,&Rx,&S);            \
   F_Mul(&Li,&Ri,&C);              \
   F_Sub(&cas2,&cas2,&Li);         \
   Li=Left[i2];                    \
   F_Add(&Left[i2],&Li,&cas2);     \
   F_Sub(&Right[i2],&Li,&cas2);    \
  }                                \
 }

/* Macro for the DiF FHT */
// C=1, S=0 N1==N2;
#define FHT_F1Butterfly(N1,N2,C,S)           \
 {FFT_DATA_TYPE D1,D2;                       \
  size_t i1=N1, i2=N2;                       \
  D1=Left[i1];D2=Left[i2];                   \
  {FFT_DATA_TYPE temp;                       \
   temp=Right[i2];                           \
   F_Add(&Left[i1],&D1,&temp);               \
   F_Sub(&Right[i2],&D1,&temp);              \
  }                                          \
 }

#define FHT_F2Butterfly(N1,N2,C,S)           \
 {FFT_DATA_TYPE D1,D2,temp;                  \
  size_t i1=N1, i2=N2;                       \
  D1=Left[i1];D2=Left[i2];                   \
   temp=Right[i1];                           \
   F_Add(&Left[i1],&D1,&temp);               \
   F_Sub(&D1,&D1,&temp);                     \
   temp=Right[i2];                           \
   F_Add(&Left[i2],&D2,&temp);               \
   F_Sub(&D2,&D2,&temp);                     \
  F_Mul(&temp,&D1,&C);                       \
  F_Mul(&Right[i1],&D2,&S);                  \
  F_Add(&Right[i1],&Right[i1],&temp);        \
  F_Mul(&temp,&D1,&S);                       \
  F_Mul(&Right[i2],&D2,&C);                  \
  F_Sub(&Right[i2],&temp,&Right[i2]);        \
 }

void
FHT_T(FFT_DATA_TYPE *Data, size_t Len)
/* Decimation in Time Hartley transform */
{
  size_t Step, Step2, Step4;
  TRIG_VARS
  size_t a,b;
  FFT_DATA_TYPE *Right,*Left;

  Step = 1;
  while (Step < Len)
    {
      Step *= 2;
      Step2 = Step/2;
      Step4 = Step/4;
      Left  = &Data[0];
      Right = &Data[Step2];

      INIT_TRIG(Step2);

      /* Do the special b=0 loop below */
      for (a = 0; a < Len; a+=Step)
         {
          FHT_T1Butterfly(a,a,1.0,0.0);
          if (Step4)
            FHT_T1Butterfly(a+Step4,a+Step4,1.0,0.0);
         }

      for (b = 1; b < Step4; b++)
        {size_t I1,I2;
         NEXT_TRIG_POW;
         for (I1=b,I2=Step2-b; I1 < Len; I1+=Step,I2+=Step)
            FHT_T2Butterfly(I1,I2,Pow_r,Pow_i);
        }
    }
}

void
FHT_F(FFT_DATA_TYPE *Data, size_t Len)
/* Decimation in frequency Hartley transform */
{
  size_t Step, Step2, Step4;
  TRIG_VARS
  size_t a,b;
  FFT_DATA_TYPE *Right,*Left;

  Step = Len;
  while (Step > 1)
    {
      Step2 = Step/2;
      Step4 = Step/4;
      Left  = &Data[0];
      Right = &Data[Step2];

      INIT_TRIG(Step2);

      /* Do the special b=0 loop below */
      for (a = 0; a < Len; a+=Step)
        {
         FHT_F1Butterfly(a,a,1.0,0.0);
         if (Step4)
           FHT_F1Butterfly(a+Step4,a+Step4,1.0,0.0);
        }

      for (b = 1; b < Step4; b++)
        {size_t I1,I2;
         NEXT_TRIG_POW;
         for (I1=b,I2=Step2-b; I1 < Len; I1+=Step,I2+=Step)
            FHT_F2Butterfly(I1,I2,Pow_r,Pow_i);
        }
      Step /= 2;
    }
}

/*
** For loading our numbers into the FFTNum array.
** Remember, this is a fixed point FFT.  We need to scale our data.
** (See Knuth, Vol2, 3rd, p367)  To allow for a max transform
** length of 2^40, we also scale it by that amount.  Since 2^40
** isn't our base, we can instead scale it by 1.1e12.  But, it's
** easier to scale by 1.0e16 because that's just two words.  Since
** the first fixed point element is the integer part, we start
** in the third element.
*/
static void
LoadNumBlockIntoNTT(size_t NX, INT32* NBuf, FFT_DATA_TYPE *FFTData)
{int z;
while (NX)
  {
   NX-=RAW_FFT_DIG;
   F_Clear(FFTData);
   for (z=0;z<RAW_FFT_DIG;z++)
     {
      FFTData->Data[3+z]=NBuf[NX+z];
     }
   FFTData++;
  }
}

static void
PutNumIntoFFTNum(FFT_DATA_TYPE *FFTNum, BigInt Num, size_t NumLen)
{size_t x,NX;
 size_t FFTLen=CalcFFTLen(NumLen);
 size_t NLen=NumLen;
 size_t FFTPos=0;
 INT32 *NBuf=(INT32*)FixedBuf;
 StartTimer(LoadTime);
 if (NumLen <8) FatalError("Too small of a FFT.\n");

  while (NLen)
    {
     NX=Min(FIXEDBUF_SIZE/sizeof(INT32),NLen);
     NLen-=NX;
     ReadNumIntoBuf(Num+NLen,NBuf,NX);
     LoadNumBlockIntoNTT(NX,NBuf,FFTNum+FFTPos);
     FFTPos+=(NX/RAW_FFT_DIG);
    }
  for (x = FFTLen/2; x < FFTLen; x++) F_Clear(&FFTNum[x]);
  StopTimer(LoadTime);
}

/*
** Do a convolution.  The data is scrambled.
*/
static void
DoScrambledFHTConv(FFT_DATA_TYPE *FFTNum1, FFT_DATA_TYPE *FFTNum2,size_t Len2)
{size_t Step=2;
 size_t Step2=Step*2;
 size_t x,y;

 while (Step < Len2)
  {
   for (x=Step,y=Step2-1;x<Step2;x+=2,y-=2)
     {FFT_DATA_TYPE h1p,h1m,h2p,h2m;
      FFT_DATA_TYPE s1,d1;
      h1p=FFTNum1[x];
      h1m=FFTNum1[y];
      F_Add(&s1,&h1p,&h1m);
      F_Sub(&d1,&h1p,&h1m);
      h2p=FFTNum2[x];
      h2m=FFTNum2[y];
      F_Mul(&h1p,&h2p,&s1);F_Mul(&h1m,&h2m,&d1);
      F_Add(&FFTNum1[x],&h1p,&h1m);
      F_DivByFloat(&FFTNum1[x],&FFTNum1[x],2.0*Len2);
      F_Mul(&h1m,&h2m,&s1);F_Mul(&h1p,&h2p,&d1);
      F_Sub(&FFTNum1[y],&h1m,&h1p);
      F_DivByFloat(&FFTNum1[y],&FFTNum1[y],2.0*Len2);
     }
   Step*=2;
   Step2*=2;
  }
 F_Mul(&FFTNum1[0],&FFTNum1[0],&FFTNum2[0]);
 F_Add(&FFTNum1[0],&FFTNum1[0],&FFTNum1[0]);
 F_DivByFloat(&FFTNum1[0],&FFTNum1[0],2.0*Len2);
 F_Mul(&FFTNum1[1],&FFTNum1[1],&FFTNum2[1]);
 F_Add(&FFTNum1[1],&FFTNum1[1],&FFTNum1[1]);
 F_DivByFloat(&FFTNum1[1],&FFTNum1[1],2.0*Len2);
}

void
DoConvolution(FFT_DATA_TYPE *FFTNum,int Cache,size_t Len2)
{
StartTimer(ConvTime);
if (Cache==-2) /* In memory convolution of both nums */
  {FFT_DATA_TYPE *FFTNum2=FFTNum+Len2;
   DumpDebug("Cm...");
   DoScrambledFHTConv(FFTNum,FFTNum2,Len2);
  }
else if (Cache==-1)  /* In memory self convolution. */
  {
   DumpDebug("C");
   DoScrambledFHTConv(FFTNum,FFTNum,Len2);
  }
#ifdef VIRTUAL_CACHE
else /* Virtual Mem based convolution. */
  {FFT_DATA_TYPE *Num1=FFTNum;
   FFT_DATA_TYPE *Num2=FFTCash[Cache].Mem;

   if (Num2==NULL)
     FatalError("Cache %d doesn't exist.\n",Cache);

   DumpDebug("C%d...",Cache);
   DoScrambledFHTConv(Num1,Num2,Len2);
  }
#else
#error
#if 0
/*
** Eyuk!  What a kludge!
*/
else
  {FILE *f;FFT_DATA_TYPE h1p,h1m;
   size_t Step=2;
   size_t Step2=Step*2;
   FFT_DATA_TYPE dn=0.5 / Len2;
   size_t x,y;
   size_t BufSize=(FIXEDBUF_SIZE/2);
   FFT_DATA_TYPE *Bufx=(FFT_DATA_TYPE*)FixedBuf;
   FFT_DATA_TYPE *Bufy=(FFT_DATA_TYPE*)(FixedBuf+BufSize);

   DumpDebug("C%d...",Cache);
   f=fopen(FFTCash[Cache].Name,"rb");
   if (f==NULL)
     FatalError("Unable to open '%s' for convolution.\n",FFTCash[Cache].Name);
   while (Step < Len2)
    {x=Step;y=Step2-1;
     while (x < Step2)
       {size_t Sz,xx,yy;
        Sz=Min(BufSize/sizeof(FFT_DATA_TYPE),Step2-x);
        fseek(f,x*sizeof(FFT_DATA_TYPE),SEEK_SET);
        fread(Bufx,sizeof(FFT_DATA_TYPE),Sz,f);
        fseek(f,(y-Sz+1)*sizeof(FFT_DATA_TYPE),SEEK_SET);
        fread(Bufy,sizeof(FFT_DATA_TYPE),Sz,f);
        xx=0;yy=Sz-1;
        while (xx < Sz)
          {FFT_DATA_TYPE h1p,h1m,h2p,h2m;
           FFT_DATA_TYPE s1,d1;
            h1p=Bufx[xx];h1m=Bufy[yy];
            s1=h1p+h1m;
            d1=h1p-h1m;
            h2p=FFTNum[x];
            h2m=FFTNum[y];
            FFTNum[x]=(h2p*s1+h2m*d1)*dn;
            FFTNum[y]=(h2m*s1-h2p*d1)*dn;
            x+=2;xx+=2;
            y-=2;yy-=2;
          }
       }
     Step*=2;
     Step2*=2;
     }
   fseek(f,0*sizeof(FFT_DATA_TYPE),SEEK_SET);fread(&h1p,sizeof(FFT_DATA_TYPE),1,f);
   fseek(f,1*sizeof(FFT_DATA_TYPE),SEEK_SET);fread(&h1m,sizeof(FFT_DATA_TYPE),1,f);
   FFTNum[0] = h1p * 2.0 * dn * FFTNum[0];
   FFTNum[1] = h1m * 2.0 * dn * FFTNum[1];
   if (ferror(f)) FatalError("Error convoluting '%s'\n",FFTCash[Cache].Name);
   fclose(f);
  }
#endif
#endif
StopTimer(ConvTime);
}


void
FwdTransform(FFT_DATA_TYPE *ddata,BigInt Num,size_t NumLen)
{
PutNumIntoFFTNum(ddata, Num, NumLen);
StartTimer(FFTTime);
FHT_F(ddata,CalcFFTLen(NumLen));
StopTimer(FFTTime);
}

void
RevTransform(FFT_DATA_TYPE *ddata,size_t NumLen)
{
StartTimer(FFTTime);
FHT_T(ddata,CalcFFTLen(NumLen));
StopTimer(FFTTime);
}

static char *PreSine[]=
  {
/* These constants were easily computed with 'BC' the simple arbitrary precision math calculator */
/* first 8 digits are integer, remainder are the fractional part. */
/* There are 150 fractional digits of trig precision.             */
/* must be multiple of 8:12345678123456781234567812345678123456781234567812345678123456781234567812345678123456781234567812345678123456781234567812345678123456781234567812345678 */
/* sin(MY_PI     ) */   "00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000",
/* sin(MY_PI/1   ) */   "00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000",
/* sin(MY_PI/2   ) */   "00000001000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000",
/* sin(MY_PI/4   ) */   "00000000707106781186547524400844362104849039284835937688474036588339868995366239231053519425193767163820786367506923115456148512462418027925368606322060",
/* sin(MY_PI/8   ) */   "00000000382683432365089771728459984030398866761344562485627041433800635627546033960089692237013785342283547148424288661493555900756010200967597920844209",
/* sin(MY_PI/16  ) */   "00000000195090322016128267848284868477022240927691617751954807754502089494763318785924580225325309234090381730992070105536611758965799834847607388826591",
/* sin(MY_PI/32  ) */   "00000000098017140329560601994195563888641845861136673167500567257264979809387302789087536807111077146318559554074206526444102156875246843024423934128599",
/* sin(MY_PI/64  ) */   "00000000049067674327418014254954976942682658314745363025752920210122553269165939563960756822220767111800723754066860977465583461318757480977938578569045",
/* sin(MY_PI/128 ) */   "00000000024541228522912288031734529459282925065466119239451477576756773788492948137394734074150581095513958936578848733233192938793411566397614206707248",
/* sin(MY_PI/256 ) */   "00000000012271538285719926079408261951003212140372319591769250038241673246747871767833029903955561663884775345994248287157242881651940232872183119599300",
/* sin(MY_PI/512 ) */   "00000000006135884649154475359640234590372580917057886317391329356732460300956624454470484805800453186249319642961151889379576099507772535469616056036418",
/* sin(MY_PI/1k  ) */   "00000000003067956762965976270145365490919842518944610213451995397146895898975797563020651153816092963359474881421583059462621697743351395786515363562739",
/* sin(MY_PI/2k  ) */   "00000000001533980186284765612303697150264079079954864575237392967614859668139754788345742540853389731269837117744742249993025966356135303008432187004097",
/* sin(MY_PI/4k  ) */   "00000000000766990318742704526938568357948576643140919452063282179694411207971791165140088172697770566664913482881640104421812106897322162628900408882419",
/* sin(MY_PI/8k  ) */   "00000000000383495187571395589072461681181381263395026034964738523091444399775767266270319103709086190051410176563723249128314730831165257377879902490999",
/* sin(MY_PI/16k ) */   "00000000000191747597310703307439909561989000933468874033859158848374994399177083027028302365730841256908624112340789899348626642531340657613250742202667",
/* sin(MY_PI/32k ) */   "00000000000095873799095977345870517210976476351187065612851145038566668164271832267303161283124134419635053559670053007705024056141865133889507065304896",
/* sin(MY_PI/64k ) */   "00000000000047936899603066884549003990494658872746866687685766923262974409624675446188513527231443676305888966195499380281183983896993808710996160087611",
/* sin(MY_PI/128k) */   "00000000000023968449808418218729186577165021820094761474895673032759838708599113797332567205509797020915643458018636747844555821565631702078127256201721",
/* sin(MY_PI/256k) */   "00000000000011984224905069706421521561596988984804731977538386699101149807274205410886123249917376072030168742683787026682948934627611987873492127200203",
/* sin(MY_PI/512k) */   "00000000000005992112452642427842879711808890861729987177878095100009358415801784637102738017589924541246727271303105350673556974334950051898642923968842",
/* sin(MY_PI/1m  ) */   "00000000000002996056226334660750454812808357059811825187868340761821634893169617065688820531362514855537392596509341664787933268279168671392277340023862",
/* sin(MY_PI/2m  ) */   "00000000000001498028113169011228854278846155361120691758586152662689052107036930298412322504276073084497835263179685168392817281736029694895241760022648",
/* sin(MY_PI/4m  ) */   "00000000000000749014056584715721130498566730655637155959302172072092993229811211254198457574008892926212986586589953514511008362975802283693809987086524",
/* sin(MY_PI/8m  ) */   "00000000000000374507028292384123903169179084633177397404762972479576692524988983484193042864693993977838867995079455437423175109176977267272428989900612",
/* sin(MY_PI/16m ) */   "00000000000000187253514146195344868824576593563617120452720982865101848665686696532504468164751264557896207528311259764288993739107502513676828203716592",
/* sin(MY_PI/32m ) */   "00000000000000093626757073098082799067286680885620193236507169473045098544368911429189871133579875295486256823279573335818448747742464160692725797898129",
/* sin(MY_PI/64m ) */   "00000000000000046813378536549092695115518138540096959503627016670088453276817707083025754660597140315261389109452177119753734957490325690527945020039902",
/* sin(MY_PI/128m) */   "00000000000000023406689268274552759505493419034844037886207223778790338459817329102580662113502601330631529438294923048956829092802590062326810383892611",
/* sin(MY_PI/256m) */   "00000000000000011703344634137277181246213503238103798093456639975905805199553603907880438579224918909780937500563357256324533766709821673664198601229328",
/* sin(MY_PI/512m) */   "00000000000000005851672317068638690809790100834139694390008505175650210891122602853089014757980725079414068659437614608263259739391872891821888149042683",
/* sin(MY_PI/1g  ) */   "00000000000000002925836158534319357928230469068955902017585715007440910152690088178818176121067494939735750970031544769680358425310995940813205116714615",
/* sin(MY_PI/2g  ) */   "00000000000000001462918079267159680529532161865963710374261522783395990017401803976143854875470026627005209923216437811182739988180440420875791104706604",
/* sin(MY_PI/4g  ) */   "00000000000000000731459039633579840460443196849417575186334531504070673169120056369712652619519689285576546904297238749954188256181388307514910183689614",
/* sin(MY_PI/8g  ) */   "00000000000000000365729519816789920254681237914263252595521492858344834989254332600184737303068848350467509909123346218497593629773412558903474446534662",
/* sin(MY_PI/16g ) */   "00000000000000000182864759908394960130398073893325934423131704687531820846300360250305278515924049446385982795912708052124276861826678426330959921496049",
/* sin(MY_PI/32g ) */   "00000000000000000091432379954197480065581218813687255727239618372000545718705809601744163644841435326940628045590202685059337143311415123644566319739369",
/* sin(MY_PI/64g ) */   "00000000000000000045716189977098740032838382140221663928079104822215218204940998727475340936728565668256705448040723282023850818695608209261878749588330",
/* sin(MY_PI/128g) */   "00000000000000000022858094988549370016425162661783086472096966705718402768595374863960508536338759137768623680542780790173852658032844780366458573716684",
/* yeah, I know we'll never go higher than this, but... <shrug> */
/* sin(MY_PI/256g) */   "00000000000000000011429047494274685008213327779850575049555660212813173264373495737003606973449346169048189640100591222777089928152100631390221370615212",
/* sin(MY_PI/512g) */   "00000000000000000005714523747137342504106757196045166501466227216186071325690292621239068825081597409215080792549087958076985350197564062716199540270797",
/* sin(MY_PI/1t  ) */   "00000000000000000002857261873568671252053390261287568122819163246886884943548529052043235524680317908007651766805772522840969225639182351691949126613162",
/* sin(MY_PI/2t  ) */   "00000000000000000001428630936784335626026696588551907170420337828294905309800169857678691125216238419260511581718240724430157498490307486093296723305064",
/* sin(MY_PI/4t  ) */   "00000000000000000000714315468392167813013348476514468973836513439753955249588885048077076943141588475480274546653389196165318539171736791879489446528151",
/* sin(MY_PI/8t  ) */   "00000000000000000000357157734196083906506674261037048910496549785577792628503528849967648556955831131729699582445455002631243662451375788818938511803545",
/* sin(MY_PI/16t ) */   "00000000000000000000178578867098041953253337133366001258195561526001498257820806037944475341114121974095145827250173000742642595455294028104730597829098",
/* sin(MY_PI/32t ) */   "00000000000000000000089289433549020976626668567038935229466191592152324373984827152536678812822932963418473641179633299864408310659049776356052532621118",
/* sin(MY_PI/64t ) */   "00000000000000000000044644716774510488313334283563959439779147149720109092693225778337155791677086296933884713033351547105533163005200276780202317013603",
/* sin(MY_PI/128t) */   "00000000000000000000022322358387255244156667141787541198020329994065547909561292826470094357824320182189648229428764005681318984699194951962086535899450",
/* sin(MY_PI/256t) */   "00000000000000000000011161179193627622078333570894465783776509549433460625182546355774077812092493925210787763016763198519293267169802771156529008276125",
/* sin(MY_PI/512t) */   "00000000000000000000005580589596813811039166785447319789984047843766816146391512700403678393801042182648033976555891792737060931030704356655581098931435",
/* What comes after 't'erabytes?  Dunno.  I'll do it alphabetically. */
/* sin(MY_PI/1u  ) */   "00000000000000000000002790294798406905519583392723670757253998055514668802420786353944521028451347165416164020868690734243673816995516939021304343425737",
/* sin(MY_PI/2u  ) */   "00000000000000000000001395147399203452759791696361836736409745794461241992363521929422223927406447987956688421411157040617571884369379900774196596492740",
/* sin(MY_PI/4u  ) */   "00000000000000000000000697573699601726379895848180918537927716243068609445075902058829298896108224580446516513074692959216974557954178931638544354766832",
/* sin(MY_PI/8u  ) */   "00000000000000000000000348786849800863189947924090459290179213539764053278649718666181358486659656231838407809989642793306638330789278812972476280632675",
/* sin(MY_PI/16u ) */   "00000000000000000000000174393424900431594973962045229647741526197160745208838830287686578362907237323255679724665613345151188431204848372638479530270167",
/* sin(MY_PI/32u ) */   "00000000000000000000000087196712450215797486981022614824202253026990212425608661513167778461705535944252230031213194882074561622368385034729089706208767",
/* sin(MY_PI/64u ) */   "00000000000000000000000043598356225107898743490511307412142562754546336190452986552749450449956280792820017870823569067722843094300937591952097646494371",
/* sin(MY_PI/128u) */   "00000000000000000000000021799178112553949371745255653706076460907404571842432575250895420379212080222758179732442247334087686628295633295103733798483823",
/* sin(MY_PI/256u) */   "00000000000000000000000010899589056276974685872627826853038877894968711389617047872262797083942970112290155995228942184162456765688330476153799742489095",
/* sin(MY_PI/512u) */   "00000000000000000000000005449794528138487342936313913426519519877642658878358618966983284403765404041340759538096042490067627886754459777010772168676508",
/* sin(MY_PI/1v  ) */   "00000000000000000000000002724897264069243671468156956713259770055091117337123071362348127934606998229291146157681829984886694282068760719883636219784541",
/* sin(MY_PI/2v  ) */   "00000000000000000000000001362448632034621835734078478356629886292079282155804505916031124683894037901206647196047441469771975320892787366191121824340327",
/* sin(MY_PI/4v  ) */   "00000000000000000000000000681224316017310917867039239178314943304106356513807624237372694931520836353938566560131639681578535499726256023807952410534885",
/* sin(MY_PI/8v  ) */   "00000000000000000000000000340612158008655458933519619589157471671811517686391983528605989039457145354105410797124844182157913395923034600571313605209975",
/* sin(MY_PI/16v ) */   "00000000000000000000000000170306079004327729466759809794578735838375551271882013190542949716440663574248447030282160586032289815161190742498857894953627",
/* sin(MY_PI/32v ) */   "00000000000000000000000000085153039502163864733379904897289367919496499689526759273551469257809343149275370146983777619019131935088122836483278092397904",
/* sin(MY_PI/64v ) */   "00000000000000000000000000042576519751081932366689952448644683959786840351461598721560733928853297994906630868968405038201621233103051360300782800989637",
/* sin(MY_PI/128v) */   "00000000000000000000000000021288259875540966183344976224322341979898243989068076746378491876920227299986935298496772643404245083718916579436062890073075",
/* sin(MY_PI/256v) */   "00000000000000000000000000010644129937770483091672488112161170989949724971201198046389011552521810937810170183486770262106205836473273313270249084999696",
/* sin(MY_PI/512v) */   "00000000000000000000000000005322064968885241545836244056080585494974937857683993982344476478018617629882172910124333519693180642132672633364823894487994",
/* sin(MY_PI/1x  ) */   "00000000000000000000000000002661032484442620772918122028040292747487478350352421361065984576729022835063222432409821258304399015055506723110676620836039",
/* sin(MY_PI/2x  ) */   "00000000000000000000000000001330516242221310386459061014020146373743740352865013726769710580579475670046878213374931064893106812354592347025654277573072",
/* sin(MY_PI/4x  ) */   "00000000000000000000000000000665258121110655193229530507010073186871870323643607244164445076816608366587847481333766950146469357345594568092453599773294",
/* sin(MY_PI/8x  ) */   "00000000000000000000000000000332629060555327596614765253505036593435935180223191169679671261724162999739474787497672679261733626883364068412300012669128",
/* sin(MY_PI/16x ) */   "00000000000000000000000000000166314530277663798307382626752518296717967592411769028289516721276563851925431274602685037872429477415331928659522279968597",
/* sin(MY_PI/32x ) */   "00000000000000000000000000000083157265138831899153691313376259148358983796493406194575968496940092219969677372408073607707597580937101233836670584252184",
/* sin(MY_PI/64x ) */   "00000000000000000000000000000041578632569415949576845656688129574179491898282643307341885515507772396735708903092378189996821255410752807878581407349066",
/* sin(MY_PI/128x) */   "00000000000000000000000000000020789316284707974788422828344064787089745949145814179927680416133601984211713228657231768267744673625136764755317794161451",
/* sin(MY_PI/256x) */   "00000000000000000000000000000010394658142353987394211414172032393544872974573468655745932415364265465336338961467496218292584599983850461446724668581717",
/* sin(MY_PI/512x) */   "00000000000000000000000000000005197329071176993697105707086016196772436487286804523595727733594315791821979774126108150916132754995563991668466255265673",
/* sin(MY_PI/1y  ) */   "00000000000000000000000000000002598664535588496848552853543008098386218243643411036263209057536180778305216173737099080679296478814087731610542077115226",
   NULL
  };

void
InitFFT(size_t Len)
/*
** If you need to use your own trig tables, you can calculate
** them here.
*/
{int x;

InitFixPoint();
x=0;while (PreSine[x]) x++;
SineTable=AlignedMalloc((x+5)*sizeof(FFT_DATA_TYPE));
if (SineTable==NULL)
  FatalError("Unable to allocate space for small initialized trig table.\n");

x=0;
while (PreSine[x])
  {
   F_SetStr(&SineTable[x],PreSine[x]);
   x++;
  }
}

void
DeInitFFT(void)
{
DeInitFixPoint();
AlignedFree(SineTable);
}



